#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import sys
import json
import socket
import sqlite3
import argparse
from os.path import join
from os.path import dirname
from contextlib import contextmanager

sys.path.insert(0, join(dirname(__file__), "../src"))

from wa_kat.settings import LOG_UDP_PORT


# Functions & classes =========================================================
@contextmanager
def create_or_open_database(filename):
    # connect database
    db = sqlite3.connect(filename)
    db.text_factory = sqlite3.OptimizedUnicode
    db.text_factory = str

    # try create tables (success only if not created from previous session)
    db.executescript("""
        CREATE TABLE IF NOT EXISTS Logs(
            id        INTEGER PRIMARY KEY AUTOINCREMENT,
            ip        VARCHAR(50),
            msg       TEXT,
            long_msg  TEXT,
            level     VARCHAR(20),
            timestamp REAL,
            url       VARCHAR(500)
        );
    """)
    db.commit()

    yield db

    db.close()


def save_to_db(db, ip, data):
    db_cur = db.cursor()
    db_cur.execute("""
        INSERT INTO Logs(
            ip,
            msg,
            long_msg,
            level,
            timestamp,
            url
        ) VALUES(?, ?, ?, ?, ?, ?)
        """,
        (
            ip,
            data["msg"],
            data["long_msg"] or "",
            data["level"],
            data["timestamp"],
            data["url"] or "",
        )
    )
    db.commit()


def address_pair(port):
    return ("0.0.0.0", port)


def address_pair_to_str(address_pair):
    return "%s:%d" % address_pair


def run_log_listenner(fn, port, quiet):
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(address_pair(port))

    def decode_data(data):
        return json.loads(data)

    with create_or_open_database(fn) as db:
        while True:
            data, address = sock.recvfrom(65535)


            try:
                save_to_db(db, address_pair_to_str(address), decode_data(data))
                if not quiet:
                    print data
            except Exception as e:
                sys.stderr.write("%s: %s\n" % (e.__class__.__name__, e))
                sys.stderr.write("Received data: %s" % data)


# Main program ================================================================
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename",
        nargs="?",
        default="logs.sqlite",
        help="Name of the SQLite database file. Default `logs.sqlite`."
    )
    parser.add_argument(
        "-p",
        "--port",
        default=LOG_UDP_PORT,
        type=int,
        help="Port for the UDP server. Default %d." % LOG_UDP_PORT
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Be quiet."
    )
    args = parser.parse_args()

    run_log_listenner(fn=args.filename, port=args.port, quiet=args.quiet)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import document


# Functions & classes =========================================================
class BlackOverlay(object):
    """
    This class is used to controll semi-transparent black layer shown at
    background of certain dialogs.
    """
    def __init__(self):
        self.el = document["black_overlay"]
        self.body_el = document.get(selector="body")[0]

    def show(self):
        self.el.style.display = "block"
        self.body_el.style.overflow = "hidden"

    def hide(self):
        self.el.style.display = "none"
        self.body_el.style.overflow = "visible"


class HelpOverlay(object):
    """
    This class binds all ``<a class="glyphicon-question-sign">`` elements, so
    they will on click show iframe overlay with linked page.

    This is used for quick help.
    """
    el = document["help_overlay"]
    black_overlay = BlackOverlay()

    @classmethod
    def show(cls, url):
        cls.el.html = "<iframe src='%s' class='help_iframe'></iframe>" % url

        cls.black_overlay.show()
        cls.el.style.display = "block"
        cls.black_overlay.el.bind("click", lambda x: cls.hide())

    @classmethod
    def hide(cls):
        cls.black_overlay.hide()
        cls.el.style.display = "none"

    @classmethod
    def bind(cls):
        def ev_handler(ev):
            ev.preventDefault()
            ev.stopPropagation()
            cls.show(ev.currentTarget.href)

        for link in document.get(selector="a.glyphicon-question-sign"):
            link.bind("click", ev_handler)


HelpOverlay.bind()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from mrc import mrc_to_marc
from mrc import item_to_mrc

from to_dc import to_dc

from iso_codes import normalize



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
This module is used to set / get values from/to almost all the inputs on the
HTML page, with exception of rules (they are set by rules_view.py).
"""
#
# Imports =====================================================================
from browser import document

from components import URLBar
from components import ISSNBar
from components import LogView
from components import UrlBoxError
from components import ISSNBoxError

from components import AuthorPickerAdapter

from components import UserKeywordHandler
from components import AlephKeywordHandler
from components import AanalysisKeywordHandler

from components import InputController

from components import ConspectDescriptor

from descriptors import StandardElDescriptor
from descriptors import AnnotationDescriptor
from descriptors import PeriodicityDescriptor


# Functions & classes =========================================================
class View(object):
    """
    View object for manpulation with the inputs on the main page.
    """
    url = StandardElDescriptor(document["url"])
    issn = StandardElDescriptor(document["issn"])
    title = StandardElDescriptor(document["title"])
    subtitle = StandardElDescriptor(document["subtitle"])
    creation_date = StandardElDescriptor(document["creation_date"])
    author = AuthorPickerAdapter  # static class
    publisher = StandardElDescriptor(document["publisher"])
    place = StandardElDescriptor(document["place"])
    language = StandardElDescriptor(document["lang"])
    conspect = ConspectDescriptor  # static class
    annotation = AnnotationDescriptor(document["annotation"])
    periodicity = PeriodicityDescriptor()

    def __init__(self):
        # all kind of progress bars and error boxes
        self.log_view = LogView
        self.url_progressbar = URLBar
        self.issn_progressbar = ISSNBar
        self.urlbox_error = UrlBoxError
        self.issnbox_error = ISSNBoxError
        self.user_kw_handler = UserKeywordHandler
        self.aleph_kw_handler = AlephKeywordHandler
        self.analysis_kw_handler = AanalysisKeywordHandler
        self.additional_info = None

        self._property_list = [
            "url",
            "issn",
            "title",
            "subtitle",
            "creation_date",
            "author",
            "publisher",
            "place",
            "language",
            "conspect",
            "annotation",
            "periodicity",
            "keywords",
            "additional_info"
        ]

    def reset_bars(self):
        """
        Set all progress bars to zero and hide them.
        """
        self.url_progressbar.reset()
        self.url_progressbar.show([0, 0])

        self.issn_progressbar.reset()

        self.urlbox_error.reset()
        self.issnbox_error.reset()

        InputController._reset_typeaheads()

    @property
    def keywords(self):
        """
        Get all keywords set by user / analysis / aleph.
        """
        dataset = [
            self.aleph_kw_handler.keywords,
            self.user_kw_handler.keywords,
            self.analysis_kw_handler.keywords,
        ]

        return sum(dataset, [])  # flattened dataset

    def get_all_properties(self):
        """
        Get dictionary with all properties readable by this class.
        """
        properties = {
            prop_name: getattr(self, prop_name)
            for prop_name in self._property_list
        }

        return properties

    def validate(self):
        """
        Validate all inputs. Highlight invalid inputs.
        """
        properties = (
            getattr(self.__class__, prop)
            for prop in self._property_list
            if hasattr(self.__class__, prop)
        )

        all_valid = True
        for prop in properties:
            if hasattr(prop, "validate"):
                # oh for fucks sake, brython has broken parser and return in
                # prop.validate() will also return from this function if used
                # with `and` operator
                result = prop.validate()
                all_valid = all_valid and result

        return all_valid

    def reset(self):
        """
        Reset all inputs back to default.
        """
        self.reset_bars()
        self.url_progressbar.reset()

        # reset all resetable components
        for prop in dir(self):
            if prop.startswith("__"):
                continue

            prop_obj = getattr(self, prop)
            if prop_obj is not None and hasattr(prop_obj, "reset"):
                prop_obj.reset()

        # reset descriptors
        properties = (
            getattr(self.__class__, prop)
            for prop in self._property_list
            if hasattr(self.__class__, prop)
        )
        for prop in properties:
            if hasattr(prop, "reset"):
                prop.reset()
            elif hasattr(prop, "__set__"):
                prop.__set__(None, "")

        self.additional_info = None


ViewController = View()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================


# Functions & classes =========================================================
class SourceString(str):
    """
    Extended ``str`` class containing also informations about source of the
    string.

    Attributes:
        val (str): Value of the string.
        source (str, default None): Source of the `val`.
    """
    def __new__(self, val, *args, **kwargs):
        try:
            return str.__new__(self, val)
        except UnicodeEncodeError:
            return str.__new__(self, val.encode("utf-8"))

    def __init__(self, val, source=None):
        self.val = val
        self.source = source

    def __repr__(self):
        if not self.source:
            return super(SourceString, self).__repr__()

        return "%s:%s" % (self.source, self.val)

    def to_dict(self):
        """
        Convert yourself to dict.
        """
        return {
            "val": self.val,
            "source": self.source,
        }



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from os.path import join

from browser import document

from .overlay_controller import BlackOverlay

# virtual filesystem / modules provided by REST API
from virtual_fs import settings


# Functions & classes =========================================================
class OutputPicker(object):
    """
    This class controlls the interface for picking the output from the MRC /
    MARC / DC conversions.

    It also allows the user to download all converted datasets as files.
    """
    el = document["output_picker"]

    overlay = BlackOverlay()

    mrc_out_el = document["mrc_output"]
    oai_out_el = document["marc_oai_output"]
    dc_out_el = document["dublin_core_output"]

    values = None
    filename = "fn"

    @classmethod
    def set(cls, values):
        """
        Set the elements from the data obtained from REST API.

        Args:
            values (dict): Dict with ``mrc``, ``oai``, ``dc`` and ``fn`` keys.
        """
        cls.mrc_out_el.text = values.get("mrc", "")
        cls.oai_out_el.text = values.get("oai", "")
        cls.dc_out_el.text = values.get("dc", "")
        cls.filename = values.get("fn", "fn")

        cls.values = values

    @classmethod
    def show(cls, values=None):
        """
        Show the interface for picking / downloading the datasets.
        """
        if values:
            cls.set(values)

        cls.el.style.display = "block"

        cls.overlay.show()
        cls.overlay.el.bind("click", lambda x: cls.hide())

    @classmethod
    def hide(cls):
        """
        Hide interface.
        """
        cls.overlay.hide()
        cls.el.style.display = "none"

    @classmethod
    def bind_download_buttons(cls):
        """
        Bind buttons to callbacks.
        """
        def on_click(ev):
            button_el = ev.target
            form_el = button_el.parent.parent.parent

            # this allows to use disabled <textearea>, which is normally not
            # sent
            content = form_el.get(selector="textarea")[0].text

            # change filename according to output
            suffix = form_el.name
            download_path = "as_file/%s.%s" % (cls.filename, suffix)
            form_el.action = join(settings.API_PATH, download_path)

            input_el = form_el.get(selector="input")[0]
            input_el.value = content

        for el in document.get(selector="button.output_download_button"):
            el.bind("click", on_click)

    @classmethod
    def reset(cls):
        """
        Reset the datasets and interface to default values.
        """
        cls.hide()

        cls.values = None
        cls.filename = "fn"
        cls.dc_out_el.text = ""
        cls.oai_out_el.text = ""
        cls.mrc_out_el.text = ""


OutputPicker.bind_download_buttons()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse informations about authors from HTML ``<meta>`` tags and
`Dublin core <http://dublincore.org>`_.
"""
#
# Imports =====================================================================
import dhtmlparser

from shared import parse_meta


# Functions & classes =========================================================
def get_html_authors(index_page):
    """
    Return list of `authors` from HTML ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "author", "HTML")


def get_dc_authors(index_page):
    """
    Return list of `authors` parsed from dublin core.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "DC.Creator", "DC")


def get_author_tags(index_page):
    """
    Parse `authors` from HTML ``<meta>`` and dublin core.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    authors = [
        get_html_authors(dom),
        get_dc_authors(dom),
    ]

    return sum(authors, [])  # return flattened list



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import sys
import argparse
from os.path import join
from os.path import dirname

from bottle import run

sys.path.insert(0, join(dirname(__file__), "../src"))
from wa_kat import rest_api  # for Bottle applications
from wa_kat import settings


# Functions & classes =========================================================
def main():
    run(
        server=settings.WEB_SERVER,
        host=settings.WEB_ADDR,
        port=settings.WEB_PORT,
        debug=settings.WEB_DEBUG,
        reloader=settings.WEB_RELOADER,
        quiet=settings.WEB_BE_QUIET,
    )


# Main program ================================================================
if __name__ == '__main__':
    # to add help with name and short description
    parser = argparse.ArgumentParser(
        description="WA-KAT server runner."
    )
    args = parser.parse_args()

    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.convertors.mrc import val_to_mrc
from wa_kat.convertors.mrc import dicts_to_mrc

from marcxml_parser import MARCXMLRecord

import pytest


# Variables ===================================================================
RECORD_EXAMPLE = """
<?xml version = "1.0" encoding = "UTF-8"?>
<record>
<metadata>
<oai_marc>
<fixfield id="LDR">-----nai-a22------a-4500</fixfield>
<fixfield id="FMT">SE</fixfield>
<fixfield id="001">web20092023182</fixfield>
<fixfield id="003">CZ-PrNK</fixfield>
<fixfield id="005">20111130083047.0</fixfield>
<fixfield id="006">m--------d--------</fixfield>
<fixfield id="007">cr-cn-</fixfield>
<fixfield id="008">091130c20099999xr--x-w-s-----0---b2cze--</fixfield>
<varfield id="015" i1=" " i2=" ">
<subfield label="a">cnb002023182</subfield>
</varfield>
<varfield id="022" i1=" " i2=" ">
<subfield label="a">1804-1051</subfield>
</varfield>
<varfield id="040" i1="1" i2="2">
<subfield label="a">ABA001</subfield>
<subfield label="b">cze</subfield>
</varfield>
</oari_marc>
</metadata>
</record>
"""


# Fixtures ====================================================================
@pytest.fixture(scope="module")
def marc_record():
    return MARCXMLRecord(RECORD_EXAMPLE)


# Functions & classes =========================================================
def test_record(marc_record):
    assert marc_record["FMT"] == "SE"


def test_val_to_mrc(marc_record):
    assert val_to_mrc("FMT", marc_record["FMT"]) == "FMT   L SE"
    assert val_to_mrc("F", marc_record["FMT"]) == "F     L SE"


def test_dicts(marc_record):
    assert dicts_to_mrc("040", marc_record["040"]) == ["04012 L $$aABA001$$bcze"]



#! /usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
from wa_kat.rest_api.to_output import _to_date_in_588


def test_url_to_fn():
    pass


def test_parse_date_range():
    pass


def test_to_date_in_588():
    assert _to_date_in_588("03.05.2017") == "3.5.2017"

    assert _to_date_in_588("2017.05.03") == "2017.5.3"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from iso3166 import countries
from iso_639_codes import translate


# Functions & classes =========================================================
def normalize(code):
    """
    Normalize language codes to ISO 639-2. If all conversions fails, return the
    `code` as it was given.

    Args:
        code (str): Language / country code.

    Returns:
        str: ISO 639-2 country code.
    """
    if len(code) == 3:
        return code

    normalized = translate(code)

    if normalized:
        return normalized

    country = countries.get(code, None)

    if country:
        return country.alpha3.lower()

    return code



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse and guess information about language of the resource. Normalize the
language tags to ISO 639-2 format.
"""
#
# Imports =====================================================================
import langdetect
import dhtmlparser

from .shared import parse_meta
from .source_string import SourceString

from ..convertors.iso_codes import normalize


# Functions & classes =========================================================
def get_html_lang_tags(index_page):
    """
    Return `languages` stored in ``<meta>`` tags.

    ``<meta http-equiv="Content-language" content="cs">`` -> ``cs``

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    lang_tag = "content-language"
    lang_tags = dom.find(
        "meta",
        fn=lambda x: x.params.get("http-equiv", "").lower() == lang_tag
    )

    return [
        SourceString(tag.params["content"], "HTML")
        for tag in lang_tags
        if "content" in tag.params
    ]


def get_html_tag_lang_params(index_page):
    """
    Parse lang and xml:lang parameters in the ``<html>`` tag.

    See
      https://www.w3.org/International/questions/qa-html-language-declarations
    for details.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    html_tag = dom.find("html")

    if not html_tag:
        return []

    html_tag = html_tag[0]

    # parse parameters
    lang = html_tag.params.get("lang")
    xml_lang = html_tag.params.get("xml:lang")

    if lang and lang == xml_lang:
        return [SourceString(lang, source="<html> tag")]

    out = []

    if lang:
        out.append(SourceString(lang, source="<html lang=..>"))

    if xml_lang:
        out.append(SourceString(xml_lang, source="<html xml:lang=..>"))

    return out


def get_dc_lang_tags(index_page):
    """
    Return `languages` stored in dublin core ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "dc.language", "DC")


def detect_language(index_page):
    """
    Detect `languages` using `langdetect` library.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        obj: One :class:`.SourceString` object.
    """
    dom = dhtmlparser.parseString(index_page)

    clean_content = dhtmlparser.removeTags(dom)

    lang = None
    try:
        lang = langdetect.detect(clean_content)
    except UnicodeDecodeError:
        lang = langdetect.detect(clean_content.decode("utf-8"))

    return SourceString(
        lang,
        source="langdetect"
    )


def get_lang_tags(index_page):
    """
    Collect informations about language of the page from HTML and Dublin core
    tags and langdetect guesses.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    lang_tags = [
        get_html_lang_tags(dom),
        get_dc_lang_tags(dom),
        [detect_language(dom)],
        get_html_tag_lang_params(dom),
    ]

    return list(sorted(set(
        SourceString(normalize(lang), source=lang.source)
        for lang in sum(lang_tags, [])
    )))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.source_string import SourceString


# Tests =======================================================================
def test_source_string():
    assert SourceString("azgabash") == "azgabash"
    assert SourceString("azgabash", source="DC") == "azgabash"

    assert "DC" in SourceString("azgabash", source="DC").__repr__()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import pytest

import requests


# Fixtures ====================================================================
@pytest.fixture
def analyze_url(web_port):
    return "http://localhost:%d/api_v1/analyze" % web_port


# Tests =======================================================================
def test_rest(analyze_url):
    resp = requests.post(analyze_url, data={"url": "http://kitakitsune.org/"})
    out = resp.json()

    assert out["status"]
    assert "error" not in out
    assert "body" in out

    assert "all_set" in out["body"]
    assert "progress" in out["body"]
    assert "values" in out["body"]


def test_rest_fail(analyze_url):
    resp = requests.post(analyze_url, data={"url": "azgabash"})
    out = resp.json()

    assert not out["status"]
    assert "error" in out
    assert out["error"]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import dhtmlparser

from source_string import SourceString


# Functions & classes =========================================================
def parse_meta(content, meta_name, source_descr, content_attr_name="content"):
    """
    Return list of strings parsed from `content` attribute from ``<meta>``
    tags with given `meta_name`.
    """
    dom = dhtmlparser.parseString(content)

    meta_tags = dom.find(
        "meta",
        fn=lambda x: x.params.get("name", "").lower() == meta_name.lower()
    )

    return [
        SourceString(tag.params[content_attr_name], source_descr)
        for tag in meta_tags
        if content_attr_name in tag.params
    ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import pytest

from wa_kat.db import request_info
from wa_kat.db.request_info import worker_mapping


# Variables ===================================================================


# Fixtures ====================================================================
@pytest.fixture
def ri_obj():
    return request_info.RequestInfo(url="http://kitakitsune.org")


# Tests =======================================================================
def test_RequestInfo(ri_obj):
    assert ri_obj.domain == "kitakitsune.org"
    assert ri_obj.creation_ts

    assert not ri_obj.downloaded_ts
    assert not ri_obj.processing_started_ts
    assert not ri_obj.processing_ended_ts

    for property_name in request_info.worker_mapping().keys():
        assert hasattr(ri_obj, property_name)

    assert not ri_obj._is_all_set()
    assert ri_obj.progress()[0] == 0

    for property_name in request_info.worker_mapping().keys():
        setattr(ri_obj, property_name, 1)

    assert ri_obj._is_all_set()
    assert ri_obj.progress()[0] == ri_obj.progress()[1]


def test_to_dict(ri_obj):
    out_dict = ri_obj.to_dict()
    keys = request_info.worker_mapping().keys()

    assert out_dict
    assert out_dict["all_set"] == False
    assert out_dict["progress"] == (0, len(keys))

    for key in keys:
        assert key in out_dict["values"]


def test_eq():
    r1 = request_info.RequestInfo(url="http://kitakitsune.org")
    r2 = request_info.RequestInfo(url="http://kitakitsune.org")
    r3 = request_info.RequestInfo(url="http://azgabash.org")

    assert r1 == r2
    assert r2 == r1

    assert r1 != r3
    assert r3 != r1

    assert r2 != r3
    assert r3 != r2


def test_lt():
    r1 = request_info.RequestInfo(url="http://kitakitsune.org")
    r2 = request_info.RequestInfo(url="http://kitakitsune.org")

    assert r1 < r2



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import bz2
import sys
import json
import os.path
import argparse
from collections import namedtuple

import retrying
import timeout_wrapper
from tqdm import tqdm
from sqlitedict import SqliteDict
from marcxml_parser import MARCXMLRecord
from edeposit.amqp.aleph import aleph
from edeposit.amqp.aleph.aleph import DocumentNotFoundException
from edeposit.amqp.aleph.aleph import InvalidAlephBaseException


# Functions & classes =========================================================
class KeywordInfo(namedtuple("KeywordInfo", ["uid",
                                             "sysno",
                                             "zahlavi",
                                             "odkazovana_forma",
                                             "angl_ekvivalent",
                                             "mdt",
                                             "mrf",
                                             "zdroj_angl_ekvivalentu",
                                             "poznamka"])):
    @classmethod
    def from_marc(cls, sysno, marc):
        def first_or_none(item):
            if not item:
                return None

            if type(item) in [tuple, list]:
                return item[0]

            return item

        def parse_mrf(mrf):
            if not mrf:
                return mrf

            return mrf.split("_")[0]

        def de_entity(s):
            if not s:
                return s

            if isinstance(s, list) or isinstance(s, tuple):
                return [de_entity(x) for x in s]

            if isinstance(s, dict):
                return {key: de_entity(val) for key, val in s.iteritems()}

            return s.replace("&nbsp;", " ") \
                    .replace("&quot;", '"') \
                    .replace("&apos;", "'") \
                    .replace("&amp;", "&")

        return cls(
            uid=first_or_none(marc["001"]),
            sysno=sysno,
            zahlavi=de_entity(first_or_none(marc["150a"])),
            odkazovana_forma=marc.get("450a", []),
            mdt=de_entity(first_or_none(marc["080a"])),
            mrf=de_entity(parse_mrf(first_or_none(marc["0802"]))),
            angl_ekvivalent=de_entity(first_or_none(marc["750a07"])),
            zdroj_angl_ekvivalentu=de_entity(first_or_none(marc["750a02"])),
            poznamka=de_entity(first_or_none(marc["680i"])),
        )


@retrying.retry(stop_max_attempt_number=3)
@timeout_wrapper.timeout(5)
def _download(doc_id):
    """
    Function used to download data authority data from aleph. Function retries
    failed attempts to download data and also has timeout.

    Args:
        doc_id (int): Sysno of the document you wish to receive.

    Returns:
        str: MARC for given record.

    Raises:
        DocumentNotFoundException: In case that document was not found.
        InvalidAlephBaseException: In case that document was not found.
    """
    return aleph.downloadMARCOAI(str(doc_id), "aut10")


def _download_items(db, last_id):
    """
    Download items from the aleph and store them in `db`. Start from `last_id`
    if specified.

    Args:
        db (obj): Dictionary-like object used as DB.
        last_id (int): Start from this id.
    """
    MAX_RETRY = 20  # how many times to try till decision that this is an end
    MAX_DOC_ID = 10000000  # this is used for download iterator

    not_found_cnt = 0  # circuit breaker
    for doc_id in xrange(last_id, MAX_DOC_ID):
        doc_id += 1
        print "Downloading %d.." % (doc_id)

        if not_found_cnt >= MAX_RETRY:
            print "It looks like this is an end:", doc_id - MAX_RETRY
            break

        try:
            record = _download(doc_id)
        except (DocumentNotFoundException, InvalidAlephBaseException):
            print "\tnot found, skipping"
            not_found_cnt += 1
            continue

        not_found_cnt = 0
        db["item_%d" % doc_id] = record
        db["last_id"] = doc_id - MAX_RETRY if doc_id > MAX_RETRY else 1

        if doc_id % 100 == 0:
            db.commit()


def download_items(cache_fn, start=None):
    """
    Open the `cache_fn` as database and download all not-yet downloaded items.

    Args:
        cache_fn (str): Path to the sqlite database. If not exists, it will be
            created.
        start (int, default None): If set, start from this sysno.
    """
    with SqliteDict(cache_fn) as db:
        last_id = db.get("last_id", 0) if not start else start
        _download_items(db, last_id)
        db.commit()


def _pick_keywords(db):
    """
    Go thru downloaded data stored in `db` and filter keywords, which are
    parsed and then yielded.

    Shows nice progress bar.

    Args:
        db (obj): Opened database connection.

    Yields:
        obj: :class:`KeywordInfo` instances for yeach keyword.
    """
    for key, val in tqdm(db.iteritems(), total=len(db)):
        # skip counter of the last downloaded document
        if key == "last_id":
            continue

        # this is optimization to speed up skipping of the unwanted elements
        # by the factor of ~20
        piece = val[:500] if len(val) > 500 else val
        if '<fixfield id="001">ph' not in piece.lower():
            continue

        parsed = MARCXMLRecord(val)

        code = parsed.get("001")
        if not code:
            continue

        # record was disabled
        if parsed["682i"]:
            continue

        if code.lower().startswith("ph"):
            yield KeywordInfo.from_marc(
                sysno=int(key.split("_")[-1]),  # item_xxx -> int(xxx)
                marc=parsed,
            )


def generate(cache_fn):
    """
    Go thru `cache_fn` and filter keywords. Store them in `keyword_list.json`.

    Args:
        cache_fn (str): Path to the file with cache.

    Returns:
        list: List of :class:`KeywordInfo` objects.
    """
    if not os.path.exists(cache_fn):
        print >> sys.stderr, "Can't access `%s`!" % cache_fn
        sys.exit(1)

    with SqliteDict(cache_fn) as db:
        for item in _pick_keywords(db):
            yield item


# Main program ================================================================
if __name__ == '__main__':
    default_cache_fn = "./aleph_kw_index.sqlite"
    default_output_fn = "./keyword_list.json"

    parser = argparse.ArgumentParser(
        description="""Aleph keyword index builder. This program may be used to
    build fast index for the keywords from AUT base."""
    )
    parser.add_argument(
        "-c",
        "--cache",
        default=default_cache_fn,
        help="Name of the cache file. Default `%s`." % default_cache_fn
    )
    parser.add_argument(
        "-o",
        "--output",
        default=default_output_fn,
        help="Name of the output file. Default `%s`." % default_output_fn
    )
    parser.add_argument(
        "-s",
        "--start-at",
        metavar="N",
        dest="start_at",
        type=int,
        help="Start from N instead of last used value."
    )
    parser.add_argument(
        "-g",
        "--generate",
        action="store_true",
        help="Don't download, only generate data from dataset."
    )

    args = parser.parse_args()

    if not args.generate:
        download_items(args.cache, start=args.start_at)

    with bz2.BZ2File(args.output + ".bz2", "w") as f:
        f.write(
            json.dumps(
                [keyword._asdict() for keyword in generate(args.cache)],
                indent=4
            )
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Functions / constants shared across the REST API.
"""
#
# Imports =====================================================================
import os
import gzip
import json
import time
import os.path
import StringIO

from bottle import request
from bottle import response
from bottle import parse_date
from bottle import HTTPResponse


# Variables ===================================================================
JSON_MIME = "application/json; charset=utf-8"


# Functions ===================================================================
def to_gzipped_file(data, out=None):
    """
    Pack `data` to GZIP and write them to `out`. If `out` is not defined,
    :mod:`stringio` is used.

    Args:
        data (obj): Any packable data (str / unicode / whatever).
        out (file, default None): Optional opened file handler.

    Returns:
        obj: File handler with packed data seeked at the beginning.
    """
    if not out:
        out = StringIO.StringIO()

    with gzip.GzipFile(fileobj=out, mode="w") as f:
        f.write(data)

    out.seek(0)
    return out


def gzipped(fn):
    """
    Decorator used to pack data returned from the Bottle function to GZIP.

    The decorator adds GZIP compression only if the browser accepts GZIP in
    it's ``Accept-Encoding`` headers. In that case, also the correct
    ``Content-Encoding`` is used.
    """
    def gzipped_wrapper(*args, **kwargs):
        accepted_encoding = request.get_header("Accept-Encoding")

        if not accepted_encoding or "gzip" not in accepted_encoding:
            return fn(*args, **kwargs)

        response.set_header("Content-Encoding", "gzip")

        return to_gzipped_file(fn(*args, **kwargs))

    return gzipped_wrapper


def gzip_cache(path):
    """
    Another GZIP handler for Bottle functions. This may be used to cache the
    files statically on the disc on given `path`.

    If the browser accepts GZIP and there is file at ``path + ".gz"``, this
    file is returned, correct headers are set (Content-Encoding, Last-Modified,
    Content-Length, Date and so on).

    If the browser doesn't accept GZIP or there is no ``.gz`` file at same
    path, normal file is returned.

    Args:
        path (str): Path to the cached file.

    Returns:
        obj: Opened file.
    """
    accept_enc = request.get_header("Accept-Encoding")

    if accept_enc and "gzip" in accept_enc and os.path.exists(path + ".gz"):
        path = path + ".gz"
        response.set_header("Content-Encoding", "gzip")

    stats = os.stat(path)

    headers = dict()
    headers['Content-Length'] = stats.st_size
    headers['Last-Modified'] = time.strftime(
        "%a, %d %b %Y %H:%M:%S GMT",
        time.gmtime(stats.st_mtime)
    )

    # I need to set `headers` dict for optional HTTPResponse use, but also set
    # hedears using `response.set_header()` for normal use
    for key, val in headers.iteritems():
        response.set_header(key, val)

    modified_since = request.environ.get('HTTP_IF_MODIFIED_SINCE')
    if modified_since:
        modified_since = parse_date(modified_since.split(";")[0].strip())

    if modified_since is not None and modified_since >= int(stats.st_mtime):
        headers['Date'] = time.strftime(
            "%a, %d %b %Y %H:%M:%S GMT",
            time.gmtime()
        )
        return HTTPResponse(status=304, **headers)

    return open(path)


def in_template_path(fn):
    """
    Return `fn` in template context, or in other words add `fn` to template
    path, so you don't need to write absolute path of `fn` in template
    directory manually.

    Args:
        fn (str): Name of the file in template dir.

    Return:
        str: Absolute path to the file.
    """
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "../templates",
        fn,
    )


def read_template(fn):
    """
    Read template `fn` from the template directory.

    Args:
        fn (str): Name of the file in template directory (relative to template
            directory).

    Returns:
        any: Returned data.
    """
    with open(in_template_path(fn)) as f:
        return f.read()


def to_json(data):
    """
    Convert `data` to pretty JSON.

    Args:
        data (any): Any JSON convertable data.

    Returns:
        unicode: Converted data.
    """
    return json.dumps(
        data,
        indent=4,
        separators=(',', ': ')
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module is used to detect titles of the webpages from several sources.
"""
#
# Imports =====================================================================
import dhtmlparser

from .shared import parse_meta
from source_string import SourceString


# Functions & classes =========================================================
def get_html_titles(index_page):
    """
    Return list of titles parsed from HTML.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    title_tags = dom.find("title")

    return [
        SourceString(tag.getContent().strip(), "HTML")
        for tag in title_tags
        if tag.getContent().strip()
    ]


def get_html_meta_titles(index_page):
    """
    Return list of titles parsed from ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "title", "Meta")


def get_dublin_core_titles(index_page):
    """
    Return list of titles parsed from dublin core inlined in ``<meta>``
    tags.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "dc.title", "DC")


def get_title_tags(index_page):
    """
    Collect data from all the functions defined in this module and return list
    of titles parsed from HTML, ``<meta>`` tags and dublin core inlined in
    ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wish to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    titles = [
        get_html_titles(dom),
        get_html_meta_titles(dom),
        get_dublin_core_titles(dom),
    ]

    return sum(titles, [])  # return flattened list



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module is used by workers launched as new processes to send back data they
analyzed in paralel.
"""
# Imports =====================================================================
import json
import time

from bottle import post
from bottle_rest import form_to_params

from ..db.request_info import RequestInfo

from ..logger import logger

from ..settings import _REQUEST_DB_SAVE


# Variables ===================================================================
global DATABASE
DATABASE = {}  #: Temporary, in-memory dict used as database.

DAY = 60 * 60 * 24  #: Amount of seconds in one day.
YEAR = DAY * 356  #: Amount of seconds in one year.


# Functions & classes =========================================================
def get_cached_or_new(url, new=False):
    """
    Look into the database and return :class:`RequestInfo` if the `url` was
    already analyzed, or create and return new instance, if not.

    If the `new` is set to True, always create new instance.

    Args:
        url (str): URL of the analyzed resource.
        new (bool, default False): Force new instance?

    Returns:
        obj: :class:`RequestInfo` instance.
    """
    garbage_collection()

    old_req = DATABASE.get(url)

    if old_req and not new:
        return old_req

    if not (url.startswith("http://") or url.startswith("https://")):
        raise ValueError("Invalid URL `%s`!" % url)

    req = RequestInfo(url=url)
    DATABASE[url] = req

    return req


def garbage_collection(time_limit=YEAR/12.0):
    """
    Collect and remove all :class:`.RequestInfo` objects older than
    `time_limit` (in seconds).

    Args:
        time_limit (float, default YEAR / 2): Collect objects older than
            this limit.
    """
    expired_request_infos = (
        ri for ri in DATABASE.values()
        if ri.creation_ts + time_limit <= time.time()
    )

    for ri in expired_request_infos:
        del DATABASE[ri.url]


# REST API ====================================================================
@post(_REQUEST_DB_SAVE)
@form_to_params
def store_property(url, property_name, value):
    """
    Look into database and store `value` under `property_name` in `url`.

    This is part of the REST API.
    """
    logger.debug(
        "store_property(): Received property_name=%s value=%s" % (
            property_name,
            value,
        ),
        url=url,
    )

    ri = get_cached_or_new(url)
    ri._set_property(property_name, json.loads(value))

    logger.info(
        "store_property(): property_name=%s saved." % (
            property_name,
        ),
        url=url,
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
REST API provider for the Aleph system.
"""
#
# Imports =====================================================================
from os.path import join

from bottle import post
from bottle_rest import form_to_params

from ..connectors import aleph
from ..settings import API_PATH


# Variables ===================================================================
API_PATH = join(API_PATH, "aleph")


# API =========================================================================
@post(join(API_PATH, "records_by_issn"))
@form_to_params
def records_by_issn(issn):
    """
    Search NTK Aleph ISSN base for given `issn`.

    Returns:
        list: List of ISSN records.
    """
    return [
        result.get_mapping()
        for result in aleph.by_issn(issn)
    ]


@post(join(API_PATH, "authors_by_name"))
@form_to_params
def authors_by_name(name):
    """
    Search NK Aleph authority base for authors matching the `name`.

    Returns:
        list: List of matching authors.
    """
    return [
        author._asdict()
        for author in aleph.Author.search_by_name(name)
    ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Connect REST API with Analyzers runner.
"""
#
# Imports =====================================================================
import traceback
from os.path import join

import requests  # for requests.exceptions.Timeout

from bottle import post
from bottle import response
from bottle_rest import form_to_params

from ..settings import API_PATH
from ..settings import REQUEST_TIMEOUT

from ..logger import logger

from shared import JSON_MIME
from db import get_cached_or_new


# REST API ====================================================================
@post(join(API_PATH, "analyze"))
@form_to_params
def analyzer_api(url):
    """
    Analyze given `url` and return output as JSON.
    """
    response.content_type = JSON_MIME

    # handle cacheing
    ri = get_cached_or_new(url)

    try:

        if ri.is_old():
            logger.info("Running the analysis.")

            # forget the old one and create new request info - this prevents
            # conflict errors
            ri = get_cached_or_new(url, new=True)

            # run the processing
            ri.paralel_processing()

    except (requests.exceptions.Timeout, requests.ConnectionError) as e:
        error_msg = """
            Požadovanou stránku {url} nebylo možné stáhnout během {timeout}
            vteřin. Zkuste URL zadat s `www.`, či zkontrolovat funkčnost
            stránek.
            <span style="display: none">{message}</span>
        """
        error_msg = error_msg.format(
            url=url,
            timeout=REQUEST_TIMEOUT,
            message=str(e.message)
        )
        logger.error(error_msg)

        return {
            "status": False,
            "log": "",  # TODO: get real log
            "error": error_msg,
        }
    except Exception as e:
        error_msg = str(e.message) + "\n" + traceback.format_exc().strip()
        logger.error(error_msg)
        return {
            "status": False,
            "log": "ri.get_log()",
            "error": error_msg,
        }

    return {
        "status": True,
        "body": ri.to_dict(),
        "log": "ri.get_log()",
    }



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import requests

from ..settings import USER_AGENT
from ..settings import REQUEST_TIMEOUT


# Functions & classes =========================================================
def download(url):
    """
    Download `url` and return it as utf-8 encoded text.

    Args:
        url (str): What should be downloaded?

    Returns:
        str: Content of the page.
    """
    headers = {"User-Agent": USER_AGENT}
    resp = requests.get(
        url,
        timeout=REQUEST_TIMEOUT,
        headers=headers,
        allow_redirects=True,
        verify=False,
    )

    def decode(st, alt_encoding=None):
        encodings = ['ascii', 'utf-8', 'iso-8859-1', 'iso-8859-15']

        if alt_encoding:
            if isinstance(alt_encoding, basestring):
                encodings.append(alt_encoding)
            else:
                encodings.extend(alt_encoding)

        for encoding in encodings:
            try:
                return st.encode(encoding).decode("utf-8")
            except UnicodeEncodeError, UnicodeDecodeError:
                pass

        raise UnicodeError('Could not find encoding.')

    return decode(resp.text, resp.encoding)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import pytest

from wa_kat.analyzers.creation_date_detector import TimeResource
from wa_kat.analyzers.creation_date_detector import mementoweb_api_tags
from wa_kat.analyzers.creation_date_detector import get_whois_tags
from wa_kat.analyzers.creation_date_detector import get_creation_date_tags


# Tests =======================================================================
def test_kitakitsune_mementoweb():
    memento_tags = mementoweb_api_tags("http://kitakitsune.org")

    assert memento_tags


def test_kitakitsune_whois():
    whois_tags = get_whois_tags("kitakitsune.org")

    assert whois_tags[0].url == "http://whois.icann.org/en/lookup?name=kitakitsune.org"
    assert whois_tags[0].date == "2009-03-27T04:04:05"
    assert whois_tags[0].val == "2009"
    assert whois_tags[0].source == "Whois"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import json
import time
import socket
import traceback

from .settings import SENTRY_DSN
from .settings import LOG_TO_SENTRY

from .settings import LOG_TO_FILE
from .settings import ERROR_LOG_PATH

from .settings import LOG_TO_STDOUT

from .settings import LOG_VIA_UDP
from .settings import LOG_UDP_ADDR
from .settings import LOG_UDP_PORT


# Functions & classes =========================================================
class Logger(object):
    def __init__(self, url=None):
        self.url = url

        self.log_path = None

        self.udp_address = None
        self.udp_port = None

        self.sentry_dsn = None

        self.should_log_to_file = False
        self.should_log_via_udp = False
        self.should_log_to_sentry = False
        self.should_log_to_stdout = True

        self.sentry_client = None

    def use_file(self, log_path):
        self.log_path = log_path
        self.should_log_to_file = True

    def use_udp(self, address, port):
        self.udp_address = address
        self.udp_port = port
        self.should_log_via_udp = True

    def use_sentry(self, sentry_dsn):
        from raven import Client
        self.sentry_client = Client(SENTRY_DSN)
        self.should_log_to_sentry = True

    def _log(self, msg, long_msg, level, url=None):
        logged_obj = {
            "msg": msg,
            "long_msg": long_msg,
            "level": level,
            "timestamp": time.time(),
            "url": url or self.url
        }

        if self.should_log_via_udp:
            self._log_to_udp(logged_obj)

        if self.should_log_to_file:
            self._log_to_file(logged_obj)

        if self.should_log_to_stdout:
            self._log_to_stdout(logged_obj)

        if self.should_log_to_sentry:
            self._log_to_sentry(logged_obj)

    @property
    def address_pair(self):
        return (self.udp_address, self.udp_port)

    def _log_to_udp(self, logged_obj):
        as_json = json.dumps(logged_obj)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sock.sendto(as_json, self.address_pair)
        except socket.gaierror as e:
            print "ERROR: Can't log to UDP logger - %s" % str(e)
        finally:
            sock.close()

    def _format_obj_to_str(self, logged_obj):
        if logged_obj["url"]:
            log_format = "{timestamp} {level} {url}; {msg}\n"
        else:
            log_format = "{timestamp} {level}; {msg}\n"

        if logged_obj["long_msg"]:
            log_format += "{long_msg}\n"

        return log_format.format(**logged_obj)

    def _log_to_file(self, logged_obj):
        with open(self.log_path, "a") as f:
            f.write(self._format_obj_to_str(logged_obj))

    def _log_to_stdout(self, logged_obj):
        print self._format_obj_to_str(logged_obj)

    def _log_to_sentry(self, logged_obj):
        if not self.sentry_client:
            print "ERROR: Requested Sentry logging, but client is not set!"
            return

        self.sentry_client.captureMessage(
            logged_obj["msg"],
            data=logged_obj,
            level=logged_obj["level"],
        )

    def emergency(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="emergency", url=url)

    def alert(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="alert", url=url)

    def critical(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="critical", url=url)

    def error(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="error", url=url)

    def warning(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="warning", url=url)

    def notice(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="notice", url=url)

    def info(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="info", url=url)

    def debug(self, msg, long_msg=None, url=None):
        self._log(msg, long_msg=long_msg, level="debug", url=url)


def set_logger_by_settingspy():
    logger = Logger()

    logger.should_log_to_stdout = LOG_TO_STDOUT

    if LOG_TO_FILE:
        logger.use_file(ERROR_LOG_PATH)

    if LOG_VIA_UDP:
        logger.use_udp(LOG_UDP_ADDR, LOG_UDP_PORT)

    if LOG_TO_SENTRY:
        logger.use_sentry(SENTRY_DSN)

    return logger


logger = set_logger_by_settingspy()


def log_exception(fn):
    def _catch_type_error_from_traceback():
        try:
            return traceback.format_exc().strip()
        except TypeError:
            return "ERROR: Can't recover traceback."

    def log_exception_decorator(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            logger.error(_catch_type_error_from_traceback())
            logger.error(
                "%s while calling %s(*args=%r, **kwargs%r): %s" % (
                    e.__class__.__name__,
                    fn.__name__,
                    args,
                    kwargs,
                    e.__str__(),
                )
            )
            raise

    return log_exception_decorator



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
import json
import time
from os.path import join

from browser import alert  # TODO: Remove
from browser import document

from view import ViewController
from rules_view import RulesViewController

from components import OutputPicker
from components import make_request
from components import func_on_enter

# virtual filesystem / modules provided by REST API
from virtual_fs import settings


# Model =======================================================================
def is_issn(s):
    """
    Does the `s` looks like valid ISSN.

    Warning:
        This function doesn't check the ISSN validity, just the basic
        differentiation between URL and ISSN:
    """
    return all([
        c.isdigit() or c in "-x "
        for c in s.lower()
    ])


class AnalysisRunnerAdapter(object):
    """
    Adapter for the analyzers API.

    This adapter is mapped to the key / button press. It then launches querying
    of the REST API in periodic interval, updating the progress bar and so on.

    It is also responsible for remapping the REST names to locally used names
    and handling of error with connection.
    """
    @classmethod
    def on_complete(cls, req):
        # handle http errors
        if not (req.status == 200 or req.status == 0):
            ViewController.log_view.add(req.text)
            ViewController.urlbox_error.show(req.text)
            return

        resp = json.loads(req.text)

        # handle structured errors
        if not resp["status"]:
            ViewController.urlbox_error.show(resp["error"])
            ViewController.log_view.add(
                "Error while running analysis: %s" % resp["error"]
            )
            return

        # keep tracking of the progress
        if not resp["body"]["all_set"]:
            ViewController.url_progressbar.show(resp["body"]["progress"])
            time.sleep(settings.GUI_TO_REST_PERIODE)
            make_request(
                url=join(settings.API_PATH, "analyze"),
                data={'url': ViewController.url},
                on_complete=cls.on_complete,
            )
            return

        # finally save the data to inputs
        ViewController.url_progressbar.show(resp["body"]["progress"])
        ViewController.log_view.add(resp["log"])

        obtained_data = json.dumps(resp["body"]["values"])
        ViewController.log_view.add("Obtained data: " + str(obtained_data))

        cls.fill_inputs(resp["body"]["values"])

    @staticmethod
    def fill_inputs(values):
        """
        Callback called when the data is received. Basically translator from
        the REST names to locally used names.
        """
        name_map = {  # TODO: get rid of this crap
            "title_tags": "title",
            "subtitle_tags": "subtitle",
            "place_tags": "place",
            "lang_tags": "language",
            "keyword_tags": "keywords",
            "author_tags": "author",
            "publisher_tags": "publisher",
            "annotation_tags": "annotation",
            "creation_dates": "creation_date",
        }

        for remote_name in values.keys():
            # special adapter for aleph keyword view
            if remote_name == "keyword_tags":
                adder = ViewController.analysis_kw_handler.add_keyword
                for keyword in values[remote_name]:
                    adder(keyword["val"])
                continue

            local_name = name_map.get(remote_name, remote_name)
            setattr(ViewController, local_name, values[remote_name])

    @classmethod
    def start(cls, ev=None):
        """
        Start the analysis.
        """
        ViewController.log_view.add("Beginning AnalysisRunner request..")

        # reset all inputs
        ViewController.reset_bars()

        # read the urlbox
        url = ViewController.url.strip()

        # make sure, that `url` was filled
        if not url:
            ViewController.urlbox_error.show("URL musí být vyplněna.")
            return

        if is_issn(url):
            ViewController.url_progressbar.hide()
            ViewController.url = ""
            ViewController.issn = url
            AlephISSNReaderAdapter.start()
            return

        ViewController.urlbox_error.hide()

        # normalize the `url`
        if not (url.startswith("http://") or url.startswith("https://")):
            url = "http://" + url
            ViewController.url = url  # store normalized url back to input

        make_request(
            url=join(settings.API_PATH, "analyze"),
            data={'url': url},
            on_complete=cls.on_complete
        )


class AlephISSNReaderAdapter(object):
    """
    Adapter for the Aleph API.

    This adapter is mapped to the key / button press. It then send the query
    to REST API and obtain data.

    It is also responsible for remapping the REST names to locally used names
    and handling of error with connection.
    """
    @classmethod
    def on_complete(cls, req):
        # handle http errors
        if not (req.status == 200 or req.status == 0):
            ViewController.log_view.add(req.text)
            ViewController.issnbox_error.show(req.text)
            ViewController.issn_progressbar.show(100)
            return

        try:
            resp = json.loads(req.text)
        except ValueError:
            resp = None

        if not resp:
            ViewController.issnbox_error.show(
                "Pro zadané ISSN nebyly nalezeny žádná data."
            )
            ViewController.log_view.add(
                "Error while reading fdata from Aleph: %s" % str(resp.text)
            )
            ViewController.issn_progressbar.show(100)
            return

        ViewController.issn_progressbar.reset()

        dataset = resp[0]
        cls._handle_aleph_keyword_view(dataset)
        AnalysisRunnerAdapter.fill_inputs(dataset)

    @staticmethod
    def _handle_aleph_keyword_view(dataset):
        """
        Translate the Aleph keywords to locally used data.
        """
        # redirect the keywords to Aleph view
        adder = ViewController.aleph_kw_handler.add_keyword
        for keyword in dataset.get("keyword_tags", []):
            adder(keyword["val"])

        if "keyword_tags" in dataset:
            del dataset["keyword_tags"]

    @classmethod
    def start(cls, ev=None):
        """
        Start the query to aleph by ISSN.
        """
        ViewController.log_view.add("Beginning AlephReader request..")

        ViewController.issnbox_error.reset()
        issn = ViewController.issn.strip()

        # make sure, that `issn` was filled
        if not issn:
            ViewController.issnbox_error.show("ISSN nebylo vyplněno!")
            ViewController.log_view.add("No ISSN! Aborting.")
            return

        ViewController.issnbox_error.hide()
        ViewController.issn_progressbar.reset()
        ViewController.issn_progressbar.show(50)
        ViewController.log_view.add("For ISSN `%s`." % issn)

        make_request(
            url=join(settings.API_PATH, "aleph/records_by_issn"),
            data={'issn': issn},
            on_complete=cls.on_complete
        )


class MARCGeneratorAdapter(object):
    """
    Adapter for the REST API for generation of the MRC / MARC / DC exports.

    This adapter is mapped to the key / button press. It then send the query
    to REST API and obtain data.

    It is also responsible for remapping the REST names to locally used names
    and handling of error with connection.
    """
    @classmethod
    def on_complete(cls, req):
        """
        Callback called when the request to REST is done. Handles the errors
        and if there is none, :class:`.OutputPicker` is shown.
        """
        # handle http errors
        if not (req.status == 200 or req.status == 0):
            ViewController.log_view.add(req.text)
            alert(req.text)  # TODO: better handling
            return

        try:
            resp = json.loads(req.text)
        except ValueError:
            resp = None

        if not resp:
            alert("Chyba při konverzi!")  # TODO: better
            ViewController.log_view.add(
                "Error while generating MARC: %s" % resp.text
            )
            return

        OutputPicker.show(resp)

    @classmethod
    def start(cls, ev):
        """
        Read all data from Views and send them to the backend.
        """
        ev.preventDefault()
        ev.stopPropagation()

        ViewController.log_view.add("Beginning MARCGenerator request..")

        if not ViewController.validate():
            ViewController.urlbox_error.show(
                "Vyplňte prosím zvýrazněné hodnoty."
            )
            ViewController.log_view.add("Aborting. Not all required data set.")
            return

        # make sure, that `url` was filled
        url = ViewController.url.strip()
        if not (url.startswith("http://") or url.startswith("https://")):
            ViewController.urlbox_error.show(
                "`%s` není validní URL! Zkuste přidat http:// | https://" % url
            )
            return

        properties = ViewController.get_all_properties()
        properties["rules"] = RulesViewController.get_dict()

        # keep the url_id from hidden input
        if "url_id" in document and document["url_id"].value:
            properties["url_id"] = document["url_id"].value

        ViewController.log_view.add(
            "Sending following data: %s" % repr(properties)
        )

        make_request(
            url=join(settings.API_PATH, "to_output"),
            data={"data": json.dumps(properties)},
            on_complete=cls.on_complete
        )


class SeederAdapter(object):
    """
    Read the data sent by Seeder. Data are included in the `seeder_data`
    element and packed in JSON.
    """
    @classmethod
    def read_data(cls):
        if not document["seeder_data"].is_set:
            return

        dataset = json.loads(document["seeder_data"].text)
        ViewController.log_view.add(
            "Obtained data from Seeder: %s" % str(dataset)
        )

        AnalysisRunnerAdapter.fill_inputs(dataset)

        if "rules" in dataset:
            RulesViewController.set_dict(dataset["rules"])


def reset_inputs(ev):
    """
    Bund the reset button to action.
    """
    ViewController.log_view.add(".reset() called.")
    ViewController.reset()


# bind buttons to actions
document["run_button"].bind("click", AnalysisRunnerAdapter.start)
document["reset_button"].bind("click", reset_inputs)
document["url"].bind("keypress", func_on_enter(AnalysisRunnerAdapter.start))
document["issn_run_button"].bind("click", AlephISSNReaderAdapter.start)
document["issn"].bind("keypress", func_on_enter(AlephISSNReaderAdapter.start))
document["marc_button"].bind("click", MARCGeneratorAdapter.start)
document["whole_brython_progress_bar"].style.display = "none"
document["content"].style.display = "block"

# read the data from the Seeder
SeederAdapter.read_data()

ViewController.log_view.add("Frontend fully loaded.")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.place_detector import get_place_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <title>HTML title</title>

    <meta name    = "geo.placename"
          content = "Praha">
</head>
<body>
Somecontent.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_place_tags():
    place_tags = get_place_tags(TEST_TEMPLATE, "nkp.cz")

    assert place_tags[0] == "Praha"
    assert place_tags[0].source == "HTML"

    assert place_tags[1] == "Narodni knihovna Ceske republiky, Klementinum 190, Praha 1, 110 00, The Czech Republic"
    assert place_tags[1].source == "Whois"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.author_detector import get_author_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <title>HTML title</title>

    <meta name    = "dc.creator"
          content = "mr. Dublin Core">

    <meta name    = "author"
          content = "mr. Meta Author">
</head>
<body>
Somecontent.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_author_tags():
    authors = get_author_tags(TEST_TEMPLATE)

    assert authors[0] == "mr. Meta Author"
    assert authors[0].source == "HTML"

    assert authors[1] == "mr. Dublin Core"
    assert authors[1].source == "DC"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module offers functions for working with MRC format used in Aleph's
desktop client.
"""
#
# Imports =====================================================================
from collections import defaultdict

from marcxml_parser import MARCXMLRecord
from marcxml_parser.tools.resorted import resorted


# Functions & classes =========================================================
def mrc_to_marc(mrc):
    """
    Convert MRC data format to MARC XML.

    Args:
        mrc (str): MRC as string.

    Returns:
        str: XML with MARC.
    """
    # ignore blank lines
    lines = [
        line
        for line in mrc.splitlines()
        if line.strip()
    ]

    def split_to_parts(lines):
        for line in lines:
            first_part, second_part = line.split(" L ", 1)

            yield line, first_part, second_part.lstrip()

    control_lines = []
    data_lines = []
    for line, first_part, second_part in split_to_parts(lines):
        if second_part.startswith("$"):
            data_lines.append(line)
        else:
            control_lines.append(line)

    # convert controlfield lines
    record = MARCXMLRecord()
    record.oai_marc = True
    for line, descr, content in split_to_parts(control_lines):
        record.controlfields[descr.strip()[:3]] = content

    def get_subfield_dict(line):
        fields = (
            (field[0], field[1:])
            for field in line.split("$$")[1:]
        )

        fields_dict = defaultdict(list)
        for key, val in fields:
            fields_dict[key].append(val)

        return fields_dict

    # convert datafield lines
    for line, descr, content_line in split_to_parts(data_lines):
        name = descr[:3]
        i1 = descr[3]
        i2 = descr[4]

        record.add_data_field(
            name,
            i1,
            i2,
            get_subfield_dict(content_line)
        )

    return record.to_XML()


def dicts_to_mrc(code, dicts):
    """
    Convert `dicts` under `code` to MRC. This is used to compose some of the
    data from user's input to MRC template, which is then converted to
    MARC XML / Dublin core.

    Args:
        code (str): Code of the aleph field, whic hshould be used.
        dicts (dict): Dict with aleph fields (i1/i2, a..z, 0..9) and
            informations in this fields.

    Returns:
        list: List of lines with MRC data.
    """
    def _dict_to_mrc(code, d):
        i1 = d.get("i1", d.get("ind1"))
        i2 = d.get("i2", d.get("ind2"))

        one_chars = [k for k in d.keys() if len(k) == 1]
        out = "%s%s%s L " % (code, i1, i2)
        for key in resorted(one_chars):
            for item in d[key]:
                out += "$$%s%s" % (key, item)

        return out

    return [
        _dict_to_mrc(code, d)
        for d in dicts
    ]


def val_to_mrc(code, val):
    """
    Convert one single `val` to MRC.

    This function may be used for control fields in MARC records.

    Args:,
        code (str): Code of the field.
        val (str): Value of the field.

    Returns:
        str: Correctly padded MRC line with field.
    """
    code = str(code)
    if len(code) < 3:
        code += (3 - len(code)) * " "

    return "%s   L %s" % (code, val)


def item_to_mrc(code, val):
    """
    Convert `val` to MRC, whether it is dict or string.

    Args:
        code (str): Code of the field.
        val (str or dict): Value of the field.

    Returns:
        list: MRC lines for output template.
    """
    if isinstance(val, basestring):
        return [val_to_mrc(code, val)]

    if isinstance(val, dict):
        val = [val]

    return dicts_to_mrc(code, val)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.language_detector import get_lang_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML lang="ch" xml:lang="pl">
<head>
    <title>HTML title</title>

    <meta name    = "DC.Language"
          content = "en">

    <meta http-equiv="Content-language"
          content="fr">
</head>
<body>
Zde je nějaký ten obsah, který by měl být rozpoznaný jako čeština. Tím myslím vážně jako čeština, což zřejmě vyžaduje poněkud víc dat, než se mi původně chtělo psát.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_lang_tags():
    lang_tags_mapping = {
        x.source: x.__str__()
        for x in get_lang_tags(TEST_TEMPLATE)
    }

    assert lang_tags_mapping["langdetect"] == "cze"

    assert lang_tags_mapping["DC"] == "eng"

    assert lang_tags_mapping["HTML"] == "fre"

    assert lang_tags_mapping["<html lang=..>"] == "cha"

    assert lang_tags_mapping["<html xml:lang=..>"] == "pol"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import pytest

from wa_kat.convertors.to_dc import to_dc


# Variables ===================================================================
TEST_DATA = {
    "publisher": u"Sociologick\xfd \xfastav AV \u010cR v.v.i.,",
    "issn": "2336-2391",
    "subtitle": "",
    "periodicity": u"\u010cast\xe9 aktualizace",
    "language": "cze",
    "author": None,
    "url": "http://dav.soc.cas.cz",
   "author": {
    "linked_forms": [
        "Grada Publishing a.s.",
        "Grada (nakladatelstv\xed)",
        "Nakladatelstv\xed Grada"
    ],
    "name": "Grada Publishing",
    "is_corporation": True,
    "record": {
        "i1": "2",
        "i2": " ",
        "a": ["Grada Publishing"],
        "b": " ",
        "7": ["kn20080316009"]
    },
    "code": "kn20080316009",
    "alt_name": "Grada Publishing [organizace] (Grada Publishing a.s., Grada (nakladatelstv\xed), Nakladatelstv\xed Grada)"
   },
    "cz_keywords": [
        {
            "zahlavi": "keyboard",
            "uid": "ph117670",
            "zdroj": "czenas"
        }, {
            "zahlavi": u"ANCA asociovan\xe1 vaskulitida",
            "uid": "ph568406",
            "zdroj": "czenas"
        }
    ],
    "conspect": {
        "ddc": "580",
        "en_name": "Plants",
        "conspect_id": "2",
        "mdt": "58",
        "name": "Botanika"
    },
    "annotation": u"N\u011bjak\xfd popisek.",
    "rules": {
        "youtube": False,
        "javascript": False,
        "calendars": False,
        "budget": 15000,
        "frequency": 365,
        "global_reject": True,
        "gentle_fetch": "default",
        "local_traps": False
    },
    "to_year": "9999",
    "place": "Praha",
    "from_year": "2007",
    "title": u"Data a v\xfdzkum - SDA Info",
    "creation_date": "2007-",
    "en_keywords": [
        {
            "zahlavi": "keyboard (musical instrument)",
            "zdroj": "eczenas"
        }, {
            "zahlavi": "ANCA-associated vasculitis",
            "zdroj": "eczenas"
        }
    ],
    "additional_info": {
        "776": u"7760  L $$tData a v\xfdzkum (Print)$$x1802-8152",
        "008": "008   L 131210c20079999xr-f||p|s||||||---b0mul-c",
        "222": u"222 0 L $$aData a v\xfdzkum - SDA Info$$b(On-line)",
        "PER": "PER   L $$a2x za rok"
    },
    "mdt": [
        {
            "mrf": "MRF",
            "mdt": "681.828: 681.816"
        }, {
            "mrf": "MRF",
            "mdt": "616.13/.14-002:616.155.3-097"
        }
    ]
}


EXPECTED_OUTPUT = u"""<?xml version="1.0" encoding="utf-8"?>
<metadata xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <dc:title>Data a výzkum - SDA Info</dc:title>
    <dc:creator id="kn20080316009">Grada Publishing</dc:creator>
    <dc:publisher>Sociologický ústav AV ČR v.v.i.,</dc:publisher>
    <dc:description>Nějaký popisek.</dc:description>
    <dc:coverage xml:lang="cze">Praha</dc:coverage>
    <dc:language schema="ISO 639-2">cze</dc:language>
    <dcterms:created>2007</dcterms:created>
    <dcterms:accrualperiodicity xml:lang="cze">Časté aktualizace</dcterms:accrualperiodicity>
    <dc:identifier rdf:resource="http://dav.soc.cas.cz"></dc:identifier>
    <dc:identifier xsi:type="ISSN">2336-2391</dc:identifier>
    <dc:identifier xsi:type="MDT">58</dc:identifier>
    <dc:identifier xsi:type="DDC">580</dc:identifier>
    <dc:subject xsi:type="dcterms:UDC">58</dc:subject>
    <dc:subject xsi:type="dcterms:DDC">580</dc:subject>
    <dc:subject xml:lang="cz">keyboard, ANCA asociovaná vaskulitida</dc:subject>
    <dc:subject xml:lang="en">keyboard (musical instrument), ANCA-associated vasculitis</dc:subject>
</metadata>
"""


# Fixtures ====================================================================
# @pytest.fixture
# def fixture():
#     pass

# with pytest.raises(Exception):
#     raise Exception()


# Tests =======================================================================
def test_conversion():
    assert to_dc(TEST_DATA).strip() == EXPECTED_OUTPUT.strip()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import time
from urlparse import urlparse
from collections import namedtuple
from multiprocessing import Process

from backports.functools_lru_cache import lru_cache

from ..logger import logger
from ..data_model import Model

from ..settings import DB_CACHE_TIME
from ..settings import DB_MAX_WAIT_TIME

from .worker import worker
from .downloader import download


# Functions ===================================================================
@lru_cache()
def worker_mapping():
    """
    Return properties from :class:`Model`, which are transported from backend
    to frontend.

    Returns:
        dict: Dict {`property_name`: `func_info`} (func_info is for worker).
    """
    return Model().analyzers_mapping().get_mapping()


# Classes =====================================================================
class Progress(namedtuple("Progress", ["done", "base"])):
    """
    Progress bar representation.

    Attr:
        done (int): How much is done.
        base (int): How much is there.
    """


# ORM =========================================================================
class RequestInfo(object):
    """
    This object is used to hold informations about requests, which are
    processed at that moment.

    The object is basically used as cache AND progress bar indicator of work.

    Attributes:
        url (str): URL to which this info object is related.
        domain (str): Domain of the URL.
        index (str): Content of the index page of URL.
        creation_ts (float): Timestamp of object creation.
        downloaded_ts (float): Timestamp showing where the :attr:`index` was
            downloaded.
        processing_started_ts (float): Timestamp showing when the processing of
            additional properties started.
        processing_ended_ts (float): Timestamp showing when the processing of
            additional properties stopped.

    Warning:
        There is more properties, one for all analyzator function from
        :mod:`.analyzers` see :func:`worker_mapping` for details.
    """
    def __init__(self, url):
        """
        Constructor.

        Args:
            url (str): URL to which this request is related.
        """
        self.url = url
        self.domain = urlparse(url).netloc
        self.index = None

        self.creation_ts = time.time()
        self.downloaded_ts = None
        self.processing_started_ts = None
        self.processing_ended_ts = None

        for key in worker_mapping().keys():
            setattr(self, key, None)

    def _set_property(self, name, value):
        """
        Set property `name` to `value`, but only if it is part of the mapping
        returned from `worker_mapping` (ie - data transported to frontend).

        This method is used from the REST API DB, so it knows what to set and
        what not, to prevent users from setting internal values.

        Args:
            name (str): Name of the property to set.
            value (obj): Any python value.

        Raises:
            KeyError: If `name` can't be set.
        """
        if name in worker_mapping().keys():
            setattr(self, name, value)
            return

        raise KeyError("Can't set `%s`!" % name)

    def _reset_set_properties(self):
        """
        Reset the progress counter back to zero.
        """
        for property_name in worker_mapping().keys():
            setattr(self, property_name, None)

    def _get_all_set_properties(self):
        """
        Collect names of set properties.

        Returns:
            set: Set containing names of all properties, which are set to \
                 non-None value.
        """
        return set(
            property_name
            for property_name in worker_mapping().keys()
            if getattr(self, property_name) is not None
        )

    def _is_all_set(self):
        """
        Is all properties set?

        Returns:
            bool: True if they are set.
        """
        return self._get_all_set_properties() == set(worker_mapping().keys())

    def progress(self):
        """
        Get progress.

        Returns:
            namedtuple: :class:`Progress`.
        """
        return Progress(
            done=len(self._get_all_set_properties()),
            base=len(worker_mapping()),
        )

    def is_old(self):
        """
        Is the object cached for too long, so it should be redownloaded?

        See :attr:`.DB_MAX_WAIT_TIME` and :attr:`.DB_CACHE_TIME` for details.

        Returns:
            bool: True if it is.
        """
        if not self.processing_started_ts:
            return True

        if self.processing_ended_ts:
            return self.processing_ended_ts + DB_CACHE_TIME < time.time()

        # in case that processing started, but didn't ended in
        # DB_MAX_WAIT_TIME
        expected_end_ts = self.creation_ts + DB_MAX_WAIT_TIME
        if expected_end_ts < time.time():
            logger.error("Prosessing timeouted and properites were not set!")

        return expected_end_ts < time.time()

    def to_dict(self):
        """
        This method is used in with connection to REST API. It basically
        converts all important properties to dictionary, which may be used by
        frontend.

        Returns:
            dict: ``{"all_set": bool, "progress": [int(done), int(how_many)], \
                  "values": {"property": [values], ..}}``
        """
        return {
            "all_set": self._is_all_set(),
            "progress": self.progress(),
            "values": {
                property_name: getattr(self, property_name) or []
                for property_name in worker_mapping().keys()
            }
        }

    def paralel_processing(self):
        """
        Lauch paralel processes (see :func:`.worker`) to asynchrnously fill
        properties with data.
        """
        self._reset_set_properties()

        self.index = download(self.url)
        self.downloaded_ts = time.time()
        self.processing_started_ts = time.time()

        # launch all workers as paralel processes
        for name, function_info in worker_mapping().iteritems():
            p = Process(
                target=worker,
                kwargs={
                    "url_key": self.url,
                    "property_name": name,
                    "function": function_info.func,
                    "function_arguments": function_info.args_func(self),
                }
            )
            p.start()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse `annotation` (short comment about page) from HTML ``<meta>`` and
`Dublin core <http://dublincore.org>`_ tags.
"""
#
# Imports =====================================================================
import dhtmlparser

from .shared import parse_meta


# Functions & classes =========================================================
def get_html_annotations(index_page):
    """
    Return `descriptions` stored in ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "description", "Meta")


def get_dc_annotations(index_page):
    """
    Return `description` stored in dublin core ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "dc.description", "DC")


def get_annotation_tags(index_page):
    """
    Return list of descriptions parsed from ``<meta>`` tags and dublin core
    inlined in ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of ``SourceString`` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    descriptions = [
        get_html_annotations(dom),
        get_dc_annotations(dom),
    ]

    return sum(descriptions, [])  # return flattened list



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse keywords from HTML meta tags and `Dublin core <http://dublincore.org>`_.
Try to analyze the text on the page and return list of tags found there.

Module maps the parserd keywords to currated keywords used in
`Aleph <http://ntk.cz>`_.
"""
#
# Imports =====================================================================
import dhtmlparser
from HTMLParser import HTMLParser

from textblob import TextBlob
from fuzzywuzzy import process

from shared import parse_meta
from source_string import SourceString

from ..rest_api.keywords import KEYWORDS
from ..rest_api.keywords import KEYWORDS_LOWER


# Functions & classes =========================================================
class MLStripper(HTMLParser):
    """
    Class used to strip HTML from tags.
    """
    def __init__(self):
        self.reset()
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return ''.join(self.fed)

    @classmethod
    def strip_tags(cls, html):
        """
        This function may be used to remove HTML tags from data.
        """
        s = cls()
        s.feed(html)

        return s.get_data()


def get_html_keywords(index_page):
    """
    Return list of `keywords` parsed from HTML ``<meta>`` tags.

    Args:
        index_page (str): Content of the page as UTF-8 string

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    keyword_lists = (
        keyword_list.split(",")
        for keyword_list in parse_meta(index_page, "keywords", "HTML")
    )

    # create SourceStrings from the list of keywords
    return [
        SourceString(keyword.strip(), source="HTML")
        for keyword in sum(keyword_lists, [])  # flattern the list
    ]


def get_dc_keywords(index_page):
    """
    Return list of `keywords` parsed from Dublin core.

    Args:
        index_page (str): Content of the page as UTF-8 string

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    keyword_lists = (
        keyword_list.split()
        for keyword_list in parse_meta(index_page, "dc.keywords", "DC")
    )

    return [
        SourceString(keyword, source="DC")
        for keyword in sum(keyword_lists, [])  # flattern the list
    ]


def try_utf_and_unicode(fn):
    def try_utf_and_unicode_decorator(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except UnicodeDecodeError as e:
            if "index_page" in kwargs:
                kwargs["index_page"] = kwargs["index_page"].decode("utf-8")
            else:
                args = list(args)
                args[0] = args[0].decode("utf-8")

            return fn(*args, **kwargs)

    return try_utf_and_unicode_decorator


@try_utf_and_unicode
def extract_keywords_from_text(index_page, no_items=5):
    """
    Try to process text on the `index_page` deduce the keywords and then try
    to match them on the Aleph's dataset.

    Function returns maximally `no_items` items, to prevent spamming the user.

    Args:
        index_page (str): Content of the page as UTF-8 string
        no_items (int, default 5): Number of items to return.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    index_page = MLStripper.strip_tags(index_page)
    tokenized_index = TextBlob(index_page).lower()

    def to_str(key):
        if isinstance(key, unicode):
            return key.encode("utf-8")

        return key

    present_keywords = [
        KEYWORDS_LOWER[key]
        for key in KEYWORDS_LOWER.keys()
        if len(key) > 3 and key in tokenized_index
    ]

    def to_source_string(key):
        source = "Keyword analysis"
        try:
            return SourceString(key, source)
        except UnicodeEncodeError:
            return SourceString(key.encode("utf-8"), source)

    multi_keywords = [
        to_source_string(key)
        for key in present_keywords
        if tokenized_index.words.count(key) >= 1
    ]

    multi_keywords = sorted(multi_keywords, key=lambda x: len(x), reverse=True)

    if len(multi_keywords) > no_items:
        return multi_keywords[:no_items]

    return multi_keywords


def get_keyword_tags(index_page, map_to_nk_set=True):
    """
    Parse `keywords` from HTML ``<meta>``, dublin core and from text.

    Args:
        index_page (str): Content of the page as UTF-8 string.
        map_to_nk_set (bool): Should the algorithm try to map keywords to
            keywords used in NK?

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    keywords = [
        get_html_keywords(dom),
        get_dc_keywords(dom),
    ]
    # do not try to match extracted_keywords, because they are based on Aleph's
    # dataset
    extracted_keywords = extract_keywords_from_text(index_page)

    keywords = sum(keywords, [])  # flattern

    if not map_to_nk_set:
        return keywords + extracted_keywords

    def try_match(keyword):
        """
        This provides chance to speed up the process a little.
        """
        kw = KEYWORDS_LOWER.get(keyword.lower())
        if kw:
            return kw

        return process.extractOne(str(keyword), KEYWORDS)[0].encode("utf-8")

    keywords = [
        SourceString(
            try_match(keyword),
            source=keyword.source,
        )
        for keyword in keywords
    ]

    return sorted(list(set(keywords + extracted_keywords)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module is used to connect and interface with the Aleph system used in NK
and NTK:

    - http://aleph.nkp.cz
    - http://aleph.techlib.cz
"""
#
# Imports =====================================================================
from collections import namedtuple

from remove_hairs import remove_hairs
from marcxml_parser import MARCXMLRecord
from edeposit.amqp.aleph import aleph

from ..settings import NTK_ALEPH_URL
from ..data_model import Model
from ..analyzers.source_string import SourceString


# Functions & classes =========================================================
def _first_or_none(array):
    """
    Pick first item from `array`, or return `None`, if there is none.
    """
    if not array:
        return None

    return array[0]


def _add_source(model):
    """
    Go over all attributes in `model` and add :class:`SourceString` to them.

    Args:
        model (obj): :class:`Model` instance.

    Returns:
        obj: :class:`Model` instance with :class:`SourceString` descriptors.
    """
    ignored_keys = {"author_tags", "original_xml", "additional_info"}

    # convert all values to source strings
    source = "Aleph"
    for key, val in model.get_mapping().iteritems():
        if key in ignored_keys:
            continue

        if type(val) in [list, tuple]:
            ss_val = [
                SourceString(item, source).to_dict()
                for item in val
            ]
        else:
            ss_val = [SourceString(val, source).to_dict()]

        setattr(model, key, ss_val)

    return model


def by_issn(issn):
    """
    Query aleph for records with given `issn`. The lookup is directed to the
    NTK's Aleph.

    Args:
        issn (str): ISSN of the periodical.

    Returns:
        obj: :class:`Model` instances for each record.
    """
    # monkeypatched to allow search in NTK's Aleph
    old_url = aleph.ALEPH_URL
    aleph.ALEPH_URL = NTK_ALEPH_URL
    records = aleph.getISSNsXML(issn, base="STK02")
    aleph.ALEPH_URL = old_url

    # process all records
    for record in records:
        marc = MARCXMLRecord(record)

        # following values were requested by @bjackova in
        # https://github.com/WebArchivCZ/WA-KAT/issues/66
        additional_info = {
            "222": marc.get("222", None),
            "PER": marc.get("PER", None),
            "776": marc.get("776", None),
            "008": marc.get("008", None),
            "alt_end_date": ""  # just reminder that it is filled later
        }
        additional_info = {
            key: val
            for key, val in additional_info.iteritems()
            if val
        }

        # check whether there is alternative date in 008
        alt_end_date = None
        alt_creation_date = None
        if additional_info["008"]:
            # 131114c20139999xr-q||p|s||||||---a0eng-c -> 2013
            alt_creation_date = additional_info["008"][7:11]

            # 131114c20139999xr-q||p|s||||||---a0eng-c -> 9999
            alt_end_date = additional_info["008"][11:15]
            if alt_end_date in ["9999", "****"]:
                alt_creation_date += "-"  # library convention is xxxx-
                alt_end_date = None

            additional_info["alt_end_date"] = alt_end_date

        # parse author
        author = Author.parse_author(marc)

        model = Model(
            url=_first_or_none(
                marc.get("856u")
            ),
            conspect=_first_or_none(
                marc.get("072a")
            ),
            annotation_tags=_first_or_none(
                marc.get("520a")
            ),
            periodicity=_first_or_none(
                marc.get("310a")
            ),
            title_tags=_first_or_none(
                marc.get("222a")
            ),
            subtitle_tags=_first_or_none(
                marc.get("245b")
            ),
            place_tags=remove_hairs(
                _first_or_none(marc.get("260a")) or ""
            ),
            author_tags=author._asdict() if author else None,
            publisher_tags=remove_hairs(
                (
                    _first_or_none(marc.get("260b")) or
                    _first_or_none(marc.get("264b")) or
                    "",
                ),
                ", "
            ),
            creation_dates=_first_or_none(
                marc.get("260c", [alt_creation_date])
            ),
            lang_tags=_first_or_none(
                marc.get("040b")
            ),
            keyword_tags=marc.get("650a07"),
            source_info=_first_or_none(
                marc.get("500a")
            ),
            original_xml=record,
            additional_info=additional_info,
        )

        yield _add_source(model)


class Author(namedtuple("Author", ["name",
                                   "code",
                                   "linked_forms",
                                   "is_corporation",
                                   "record",
                                   "alt_name"])):
    """
    Class for representing author metadata.

    Attributes:
        name (str): Name of the author.
        code (str): Authority code of the author.
        linked_forms (list): List of strings with alternative names.
        is_corporation (bool): True if the record is for the corporation.
        record (str): All author fields from ``100`` / ``110``.
        alt_name (str): Alternative name composed from `name` and
            `linked_forms`.
    """
    @classmethod
    def parse_author(cls, marc):
        """
        Parse author from `marc` data.

        Args:
            marc (obj): :class:`.MARCXMLRecord` instance. See module
                :mod:`.marcxml_parser` for details.

        Returns:
            obj: :class:`Author`.
        """
        name = None
        code = None
        linked_forms = None
        is_corporation = None
        record = None

        # parse informations from the record
        if marc["100a"]:  # persons
            name = _first_or_none(marc["100a"])
            code = _first_or_none(marc["1007"])
            is_corporation = False
            record = marc.datafields["100"][0]  # transport all fields
        elif marc["110a"]:  # corporations
            name = _first_or_none(marc["110a"])
            code = _first_or_none(marc["1107"])
            linked_forms = marc["410a2 "]
            is_corporation = True
            record = marc.datafields["110"][0]  # transport all fields
        else:
            return None

        # parse linked forms (alternative names)
        linked_forms = marc["410a2 "]

        # put together alt_name
        type_descriptor = ["osoba", "organizace"]
        alt_name = "%s [%s]" % (name, type_descriptor[is_corporation])
        if linked_forms:
            alt_name += " (" + ", ".join(linked_forms) + ")"

        return cls(
            name=name,
            code=code,
            linked_forms=linked_forms,
            is_corporation=is_corporation,
            record=record,
            alt_name=alt_name,
        )

    @classmethod
    def search_by_name(cls, name):
        """
        Look for author in NK Aleph authority base by `name`.

        Args:
            name (str): Author's name.

        Yields:
            obj: :class:`Author` instances.
        """
        records = aleph.downloadRecords(
            aleph.searchInAleph("aut", name, False, "wau")
        )

        for record in records:
            marc = MARCXMLRecord(record)

            author = cls.parse_author(marc)

            if author:
                yield author



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Worker definition for analyzers runner. Functions defined there are run as
standalone processes.
"""
#
# Imports =====================================================================
import json
import traceback

import requests

from ..logger import logger

from ..settings import _WEB_URL
from ..settings import _REQUEST_DB_SAVE

from ..settings import REQUEST_TIMEOUT


# Functions & classes =========================================================
def _save_to_database(url, property_name, data):
    """
    Store `data` under `property_name` in the `url` key in REST API DB.

    Args:
        url (obj): URL of the resource to which `property_name` will be stored.
        property_name (str): Name of the property under which the `data` will
            be stored.
        data (obj): Any object.
    """
    data = json.dumps([
        d.to_dict() if hasattr(d, "to_dict") else d
        for d in data
    ])

    logger.debug("_save_to_database() data: %s" % repr(data))

    requests.post(
        _WEB_URL + _REQUEST_DB_SAVE,
        timeout=REQUEST_TIMEOUT,
        allow_redirects=True,
        verify=False,
        data={
            "url": url,
            "value": data,
            "property_name": property_name,
        }
    )

    logger.info(
        "`%s` for `%s` sent to REST DB." % (
            property_name,
            url,
        )
    )


def worker(url_key, property_name, function, function_arguments):
    """
    This function usually runs as process on the background.

    It runs ``function(*function_arguments)`` and then stores them in REST API
    storage.

    Warning:
        This function puts data into DB, isntead of returning them.

    Args:
        url_key (str): Key which will be used for database lookup.
        property_name (str): Name of the property used to store data.
        function (obj): Function used to load the data.
        function_arguments (list): List of parameters for function which will
            be called to retreive data.
        error_log_path (str): If set, log errors into this file, otherwise
            stderr.
    """
    # this may take some time, hence outside transaction manager
    error_msg = None
    try:
        data = function(*function_arguments)
    except Exception as e:
        data = []
        error_msg = "Error: " + traceback.format_exc().strip()
        error_msg += "\n" + str(e.message)

    # handle logging of possible error message
    if error_msg:
        logger.error(error_msg)
        error_msg = None

    func_name = str(function.__name__)
    logger.info(
        "Attempting to save output from `%s`." % func_name
    )

    # save `data` into RequestInfo object property
    return _save_to_database(
        url=url_key,
        property_name=property_name,
        data=data
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.title_detector import get_title_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <title>HTML title</title>

    <meta name    = "DC.Title"
          lang    = "en"
          content = "Dublin Core title">

    <meta name    = "DC.Title"
          lang    = "cs"
          content = "Dublin Core titulek">

    <meta name    = "title"
          content = "<meta> title">
</head>
<body>
Somecontent.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_titles():
    titles = get_title_tags(TEST_TEMPLATE)

    assert titles[0] == "HTML title"
    assert titles[0].source == "HTML"

    assert titles[1] == "<meta> title"
    assert titles[1].source == "Meta"

    assert titles[2] == "Dublin Core title"
    assert titles[2].source == "DC"
    assert titles[3] == "Dublin Core titulek"
    assert titles[3].source == "DC"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module handles list of Aleph's keywords, provides REST API with this
keywords and allows translation of those keywords to their codes and so on.

This module cointains four major variables:

    - KW_DICT (dict): ``keyword: {info}`` mapping.
    - KEYWORDS (list): List of keywords.
    - KEYWORDS_LOWER (list): List of keywords.lower()
    - KW_CACHE_PATH (str): Path to the keyword file in ``/tmp`` (bottle
      optimization).
"""
#
# Imports =====================================================================
import bz2
import json
import os.path
from os.path import join
from collections import OrderedDict

from bottle import get
from bottle import response

from ..settings import API_PATH

from shared import gzip_cache
from shared import JSON_MIME
from shared import to_gzipped_file


# Loaders =====================================================================
def read_kw_file():
    """
    Read content of the file containing keyword informations in JSON. File is
    packed using BZIP.

    Returns:
        list: List of dictionaries containing keywords.
    """
    self_path = os.path.dirname(__file__)
    kw_list_path = join(self_path, "../templates/keyword_list.json.bz2")

    with bz2.BZ2File(kw_list_path) as f:
        kw_list = f.read()

    return json.loads(kw_list)


def build_kw_dict(kw_list):
    """
    Build keyword dictionary from raw keyword data. Ignore invalid or
    invalidated records.

    Args:
        kw_list (list): List of dicts from :func:`read_kw_file`.

    Returns:
        OrderedDict: dictionary with keyword data.
    """
    kw_dict = OrderedDict()
    sorted_list = sorted(
        kw_list,
        key=lambda x: x.get("zahlavi").encode("utf-8")
    )

    for keyword_data in sorted_list:
        if "zahlavi" not in keyword_data:
            continue

        zahlavi = keyword_data["zahlavi"].encode("utf-8")
        old_record = kw_dict.get(zahlavi)

        if not old_record:
            kw_dict[zahlavi] = keyword_data
            continue

        key = "angl_ekvivalent"
        if not old_record.get(key) and keyword_data.get(key):
            kw_dict[zahlavi] = keyword_data
            continue

        key = "zdroj_angl_ekvivalentu"
        if not old_record.get(key) and keyword_data.get(key):
            kw_dict[zahlavi] = keyword_data
            continue

        if len(str(keyword_data)) > len(str(old_record)):
            kw_dict[zahlavi] = keyword_data
            continue

    return kw_dict


# Variables ===================================================================
_INITIALIZED = False
KW_DICT = None  # Output from :func:`build_kw_dict`.
KEYWORDS = None  # List of strings with keywords.
KEYWORDS_LOWER = None  # List of strings with keywords.lower()

#: Path to the unpacked keyword list in /tmp. This is used as bottle
#: optimization.
KW_CACHE_PATH = None


def init():
    """
    Initialize all global variables (:attr:`.KW_DICT`, :attr:`.KEYWORDS`,
    :attr:`.KEYWORDS_LOWER`, :attr:`.KW_CACHE_PATH`) to their values.

    Global variables are then used from analyzers and so on.
    """
    global _INITIALIZED

    if _INITIALIZED:
        return

    global KW_DICT
    global KEYWORDS
    global KW_CACHE_PATH
    global KEYWORDS_LOWER

    KW_DICT = build_kw_dict(read_kw_file())
    KEYWORDS = sorted([k.decode("utf-8") for k in KW_DICT.keys()])
    KEYWORDS_LOWER = {
        k.lower(): k
        for k in KEYWORDS
    }
    keywords_json = json.dumps(KEYWORDS)
    KW_CACHE_PATH = "/tmp/wa_kat_cache_keywords.json"

    # create cached files
    with open(KW_CACHE_PATH, "w") as f:
        f.write(keywords_json)
    with open(KW_CACHE_PATH + ".gz", "w") as f:
        to_gzipped_file(keywords_json, out=f)

    _INITIALIZED = True


init()


# Functions ===================================================================
def keyword_to_info(keyword):
    """
    Get keyword dict based on the `keyword`.

    Args:
        keyword (str): Keyword as string.

    Returns:
        dict: Additional keyword info.
    """
    return KW_DICT.get(keyword)


# API =========================================================================
@get(join(API_PATH, "kw_list.json"))
def get_kw_list():
    """
    Virtual ``kw_list.json`` file.

    List of all keywords on one JSON page. This is later used by the typeahead
    script, which shows keyword hints to user.
    """
    response.content_type = JSON_MIME

    return gzip_cache(KW_CACHE_PATH)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import sys
import bz2
import json
import os.path
import argparse

import marcxml_parser

from tqdm import tqdm


# Functions & classes =========================================================
def read_file(fn):
    with bz2.BZ2File(fn) as f:
        data = f.read().split("</record>")

    for chunk in tqdm(data, total=len(data)):
        yield chunk + "\n</record>\n"


def records_iterator(xml_records):
    for record in xml_records:
        if "<record" not in record:
            continue

        yield marcxml_parser.MARCXMLRecord(record)


def parse_conspectus(records):
    conspectus = {}

    conspects = []
    subconspects = []
    for record in records:
        if record.get("190k"):
            conspects.append(record)
        else:
            subconspects.append(record)

    for conspect in conspects:
        uid = conspect["190k"][0]
        name = conspect["190x"][0]

        if name.startswith("*"):
            name = name[1:]

        conspectus[uid] = {
            "id": uid,
            "name": name,
            "subconspects": {}
        }

    def parse_subconspect(record):
        uid = record["1909"][0]
        name = record["190x"][0]
        mdt = record["190a"][0]

        en_name = (record.get("790x") or [None])[0]
        ddc = (record.get("790a") or [None])[0]

        return uid, mdt, {
            "conspect_id": uid,
            "name": name,
            "mdt": mdt,
            "en_name": en_name,
            "ddc": ddc,
        }

    for subconspect in subconspects:
        uid, mdt, record = parse_subconspect(subconspect)

        conspectus[uid]["subconspects"][mdt] = record

    return conspectus


# Main program ================================================================
if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="""
            This program may be used to convert Conspectus / Subconspectus set
            in MARC XML to JSON."""
    )
    parser.add_argument("XML_FILE", help="MARC XML file packed in .bz2.")
    args = parser.parse_args()

    if not os.path.exists(args.XML_FILE):
        sys.stderr.write("`%s` not found!\n" % args.XML_FILE)
        sys.exit(1)

    if not args.XML_FILE.endswith(".bz2"):
        sys.stderr.write("File is required to be XML packed in .bz2.")
        sys.exit(1)

    records = records_iterator(
        read_file(args.XML_FILE)
    )
    print json.dumps(
        parse_conspectus(records),
        indent=4,
        sort_keys=True,
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
This module is used to set / get values from/to Rules section on the HTML page.
"""
#
# Imports =====================================================================
from browser import document

from descriptors import RadioDescriptor
from descriptors import StandardElDescriptor


# Functions & classes =========================================================
class RulesView(object):
    """
    View object for manipulation with rules on the HTML page.
    """
    budget = RadioDescriptor("budget_radio", val_type=int)
    gentle_fetch = RadioDescriptor("gentle_fetch_radio", val_type=str)
    local_traps = RadioDescriptor("local_traps_radio")
    calendars = RadioDescriptor("calendars_radio")
    youtube = RadioDescriptor("youtube_radio")
    global_reject = RadioDescriptor("global_reject_radio")
    javascript = RadioDescriptor("javascript_radio")
    frequency = StandardElDescriptor(document["freq"])

    def __init__(self):
        self._property_names = [
            "budget",
            "gentle_fetch",
            "local_traps",
            "calendars",
            "youtube",
            "global_reject",
            "javascript",
            "frequency",
        ]

    def get_dict(self):
        """
        Convert all rules to dict and return them.
        """
        out = {
            property_name: getattr(self, property_name)
            for property_name in self._property_names
        }

        if "frequency" in out:
            out["frequency"] = int(out["frequency"])

        return out

    def set_dict(self, incomming):
        """
        Set all rules from the `incomming` dictionary.
        """
        for key, val in incomming.items():
            if val and key in self._property_names:
                setattr(self, key, val)


RulesViewController = RulesView()
